/* Copyright(C) 2004 Hiroyuki OYAMA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_request.h"
#include "http_main.h"
#include "ap_config.h"

#include <sys/types.h>


#include "http_core.h"
#include "util_date.h"
#include "http_conf_globals.h"
#include <stdarg.h>

#include "config.h"

#ifdef USE_MMAP_FILES
#include <sys/mman.h>

#ifndef MMAP_THRESHOLD
#ifdef SUNOS4
#define MMAP_THRESHOLD          (8*1024)
#else
#define MMAP_THRESHOLD          1
#endif  /* SUNOS4 */
#endif  /* MMAP_THRESHOLD */

#ifndef MMAP_LIMIT
#define MMAP_LIMIT              (4*1024*1024)
#endif  /* MMAP_LIMIT */

struct mmap_rec {
    void *mm;
    size_t length;
};

static void mmap_cleanup(void *mmv)
{
    struct mmap_rec *mmd = mmv;

    if (munmap(mmd->mm, mmd->length) == -1) {
        ap_log_error(APLOG_MARK, APLOG_ERR, NULL,
                     "Failed to munmap memory of length %ld at 0x%lx",
                     (long)mmd->length, (long)mmd->mm);
    }
}
#endif  /* USE_MMAP_FILES */


module MODULE_VAR_EXPORT ezweb_download_module;

typedef struct {
    table *type;
    const char *confirm_msg_path;
    const char *success_msg_path;
    const char *fail_msg_path;
} ezweb_download_conf;


struct hdml_download_rec {
    long offset;
    long count;
    const char *name;
};

struct hdml_device_dnld_rec {
    const char *url;
    const char *name;
    size_t size;
    const char *disposition;
    const char *title;
};

static const char CONTENT_TYPE_X_UP_DOWNLOAD[] = "application/x-up-download";
static const char CONTENT_TYPE_X_HDML[]        = "text/x-hdml; charset=Shift_JIS";
static const char HDML_CONFIRM[] =
    "<hdml version=3.0 ttl=\"0\" public=true>\n"
    "<choice>\n"
    "<ce task=\"GO\" dest=\"device:data/dnld?url=%s&name=%s&size=%ld&disposition=%s&title=%s\" label=\"GET\">\x83\x5f\x83\x45\x83\x93\x83\x8d\x81\x5b\x83\x68\n"
    "<ce task=\"PREV\" label=\"Back\">\x82\xe0\x82\xc7\x82\xe9\n"
    "</choice>\n"
    "</hdml>\n";
static const char HDML_SUCCESS[] =
    "<hdml version=3.0 ttl=\"0\" public=true>\n"
    "<display>\n"
    "<action type=accept task=return>\n"
    "<wrap>\x83\x5f\x83\x45\x83\x93\x83\x8d\x81\x5b\x83\x68\x82\xaa\x8a\xae\x97\xb9\x82\xb5\x82\xdc\x82\xb5\x82\xbd\n"
    "</display>\n"
    "</hdml>\n";
static const char HDML_FAIL[] =
    "<hdml version=3.0 ttl=\"0\" public=true>\n"
    "<display>\n"
    "<action type=accept task=return>\n"
    "<wrap>\x83\x5f\x83\x45\x83\x93\x83\x8d\x81\x5b\x83\x68\x82\xc9\x8e\xb8\x94\x73\x82\xb5\x82\xdc\x82\xb5\x82\xbd\n"
    "</display>\n"
    "</hdml>\n";

typedef struct {
    const char *key;
    const char *value;
} hdml_data_type_rec;
static const hdml_data_type_rec DISPOSITION_TABLE[] = {
    { "bmp", "devabm" }, // 1bit image
//  { "bmp", "dev2abm" }, // 2bit image
    { "png", "dev8aww" }, // background image
//  { "png", "devfr0i" }, // photo frame image
//  { "png", "devst0j" }, // photo stamp image
    { "jpg", "devjaww" }, // background image
    { "jpeg", "devjaww" },
    { "gif", "devgi0z" }, // background image
//  { "gif", "devga2z" }, // screen saver image
    { "qcp", "dev13anq" }, // melody
    { "pmd", "devcman" },
//  { "pmd", "dev3anp" }, // melody
    { "pmd3", "dev3anp" },
    { "mmf", "devmfan" },
//  { "mmf", "dev4anm" },
//  { "mmf", "devm39z" },
//  { "mmf", "devm53z" },
    { "kjx", "devkdjx" },
//  { "kjx", "devkdj2" },
    { "vcf", "devvcf" },
    { "vcs", "devvcs" },
    { NULL, NULL }
};
static unsigned short CRC_INIT = 0xffff;
static unsigned short CRC_TABLE[] = {
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
    0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
    0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
    0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
    0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
    0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
    0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
    0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
    0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
    0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
    0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
    0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
    0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
    0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
    0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
    0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
    0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
    0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
    0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
    0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
    0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
    0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
    0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
};
static int CRC_LENGTH = 2;


static void init_module(server_rec *s, pool *p)
{
    ap_add_version_component(MOD_EZWEB_DOWNLOAD_VERSION_STRING);
}

static void *create_dir_conf(pool *p, char *path)
{
    ezweb_download_conf *conf;

    conf = ap_pcalloc(p, sizeof(ezweb_download_conf));
    conf->type = ap_make_table(p, 0);
    conf->confirm_msg_path = NULL;
    conf->success_msg_path = NULL;
    conf->fail_msg_path    = NULL;

    return (void *)conf;
}

static const char *cmd_add_datatype(cmd_parms *cmd, ezweb_download_conf *conf,
                                    char *datatype, char *ext)
{
    if (*ext == '.')
        ++ext;
    ap_table_setn(conf->type, ext, datatype);
    return NULL;
}

static void plustospace(char *str)
{
    register int x;
    for(x=0;str[x];x++) if(str[x] == '+') str[x] = ' ';
}

static table *parse_request_param(request_rec *r)
{
    table *param = ap_make_table(r->pool, 3);
    const char *args, *chunk;

    args = r->args ? r->args : "";
    while (*args && (chunk = ap_getword(r->pool, &args, '&'))) {
        char *key, *value;

        key = ap_getword(r->pool, &chunk, '=');
        value = ap_getword(r->pool, &chunk, '=');
        plustospace(key);
        plustospace(value);
        ap_unescape_url(key);
        ap_unescape_url(value);
        ap_table_set(param, key, value);
    }

    return param;
}


static const char *find_filename(pool *p, ezweb_download_conf *conf,
                                 const char *path)
{
    char *slash;

    if ((slash = strrchr(path, '/')) == NULL)
        return "Unknown";
    return (const char *)ap_pstrdup(p, slash + 1);
}

static const char *find_disposition(pool *p, ezweb_download_conf *conf,
                                    const char *path)
{
    char *ext;
    const char *disposition;
    const hdml_data_type_rec *default_type = DISPOSITION_TABLE;

    if ((ext = strrchr(path, '.')) == NULL)
        return "";
    ++ext;

    /* find user defined datatype */
    disposition = ap_table_get(conf->type, ext);
    if (disposition != NULL)
        return (const char *)ap_pstrdup(p, disposition);

    /* find default datatype */
    for (; default_type->key != NULL; default_type++) {
        if (strcmp(ext, default_type->key) != 0)
            continue;
        return (const char *)ap_pstrdup(p, default_type->value);
    }
 
    return "";
}

static void ed_update_crc(unsigned short *crc, char *src, int length)
{
    int i;

    for (i = 0; i < length; i++) {
        unsigned char s = src[i];
        *crc = CRC_TABLE[(*crc >> 8 ^ s) & 0xff] ^ (*crc << 8);
    }
}

static size_t ed_send_fd(FILE *f, request_rec *r, unsigned short *crc)
{
    char buff[IOBUFSIZE];
    size_t total_bytes_sent = 0;
    register int n, w, o, len;

    ap_soft_timeout("send body", r);
    while (!r->connection->aborted) {
        len = IOBUFSIZE;
        /* read file */
        while ((n = fread(buff, sizeof(char), len, f)) < 1
               && ferror(f) && errno == EINTR && !r->connection->aborted)
            continue;

        if (n < 1)
            break;
        o = 0;
        ed_update_crc(crc, buff, len);

        /* send file */
        while (n && !r->connection->aborted) {
            w = ap_bwrite(r->connection->client, &buff[o], n);
            if (w > 0) {
                ap_reset_timeout(r);
                total_bytes_sent += w;
                n -= w;
                o += w;
            }
            else if (w < 0) {
                if (!r->connection->aborted) {
                    ap_log_rerror(APLOG_MARK, APLOG_INFO, r,
                                  "client stopped connection before send body completed");
                    ap_bsetflag(r->connection->client, B_EOUT, 1);
                    r->connection->aborted = 1;
                }
                break;
            }
        }
    }

    ap_kill_timeout(r);
    if (r->sent_bodyct)
        ap_bgetopt(r->connection->client, BO_BYTECT, &r->bytes_sent);

    return total_bytes_sent;
}

#ifdef USE_MMAP_FILES
#ifndef MMAP_SEGMENT_SIZE
#define MMAP_SEGMENT_SIZE 32768
#endif  /* MMAP_SEGMENT_SIZE */

static size_t ed_send_mmap(void *mm, request_rec *r, unsigned short *crc)
{
    size_t length = r->finfo.st_size;
    size_t offset = 0;
    size_t total_bytes_sent = 0;
    int n, w;

    if (length == 0)
        return 0;

    ap_soft_timeout("send mmap", r);
    while (!r->connection->aborted && offset < length) {
        if (length - offset > MMAP_SEGMENT_SIZE) {
            n = MMAP_SEGMENT_SIZE;
        }
        else {
            n = length - offset;
        }

        while (n && !r->connection->aborted) {
            ed_update_crc(crc, (char *)mm + offset, n);
            w = ap_bwrite(r->connection->client, (char *)mm + offset, n);
            if (w > 0) {
                ap_reset_timeout(r);
                total_bytes_sent += w;
                n -= w;
                offset += w;
            }
            else if (w < 0) {
                if (!r->connection->aborted) {
                    ap_log_rerror(APLOG_MARK, APLOG_INFO, r,
                                  "client stopped connection before send mmap completed");
                    ap_bsetflag(r->connection->client, B_EOUT, 1);
                    r->connection->aborted = 1;
                }
                break;
            }

        }

    }

    ap_kill_timeout(r);
    if (r->sent_bodyct)
        ap_bgetopt(r->connection->client, BO_BYTECT, &r->bytes_sent);

    return total_bytes_sent;
}
#endif  /* USE_MMAP_FILES */

static int ed_send_crc(unsigned short crc, request_rec *r)
{
    unsigned char buff[CRC_LENGTH];

    buff[0] = (crc >> 8) & 0xff;
    buff[1] = crc & 0xff;
    ap_rwrite((const void *)buff, CRC_LENGTH, r);

    return OK;
}

static int state_confirm(request_rec *r, ezweb_download_conf *conf, table *param)
{
    struct hdml_device_dnld_rec dnld_rec;
    
    dnld_rec.url = (const char *)ap_construct_url(r->pool, r->uri, r);
    dnld_rec.name = find_filename(r->pool, conf, r->filename);
    dnld_rec.size = r->finfo.st_size + CRC_LENGTH;
    dnld_rec.disposition = find_disposition(r->pool, conf, r->filename);
    dnld_rec.title = dnld_rec.name;

    r->content_type = CONTENT_TYPE_X_HDML;
    ap_send_http_header(r);
    ap_rprintf(r, HDML_CONFIRM,
               dnld_rec.url, dnld_rec.name, dnld_rec.size,
               dnld_rec.disposition, dnld_rec.title);

    return OK;
}

static int state_download(request_rec *r, ezweb_download_conf *conf, table *param)
{
    int rc;
    FILE *fp;
#ifdef USE_MMAP_FILES
    caddr_t mm;
#endif
    unsigned short crc = CRC_INIT;

    fp = ap_pfopen(r->pool, r->filename, "r");
    if (fp == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                      "file permissions deny server access: %s", r->filename);
        return FORBIDDEN;
    } 

    ap_update_mtime(r, r->finfo.st_mtime);
    ap_set_last_modified(r);
    ap_set_etag(r);
    rc = ap_meets_conditions(r);
    if ((rc != OK) || (rc = ap_set_content_length(r, r->finfo.st_size + CRC_LENGTH)))
        return rc;

    r->content_type = CONTENT_TYPE_X_UP_DOWNLOAD;

#ifdef USE_MMAP_FILES
    ap_block_alarms();
    if ((r->finfo.st_size >= MMAP_THRESHOLD)
        && (r->finfo.st_size < MMAP_LIMIT)
        && !r->header_only)
    {
        mm = mmap(NULL, r->finfo.st_size, PROT_READ, MAP_PRIVATE, fileno(fp), 0);
        if (mm == (caddr_t) - 1)
            ap_log_rerror(APLOG_MARK, APLOG_CRIT, r,
                          "ezweb-download: mmap failed: %s", r->filename);
    }
    else {
        mm = (caddr_t) - 1;
    }

    if (mm == (caddr_t) - 1) {
        ap_unblock_alarms();
#endif

        ap_send_http_header(r);
        if (!r->header_only) {
            /* send file by stdio */
           ed_send_fd(fp, r, &crc);
           ed_send_crc(crc, r);
        }
#ifdef USE_MMAP_FILES
    }
    else {
        struct mmap_rec *mmd;

        /* send file by mmap */
        mmd = ap_palloc(r->pool, sizeof(*mmd));
        mmd->mm = mm;
        mmd->length = r->finfo.st_size;
        ap_register_cleanup(r->pool, (void *)mmd, mmap_cleanup, mmap_cleanup);
        ap_unblock_alarms();
        ap_send_http_header(r);
        if (!r->header_only) {
            ed_send_mmap(mm, r, &crc);
            ed_send_crc(crc, r);
        }
    }
#endif

    ap_pfclose(r->pool, fp);

    return OK;
}


static int state_empty(request_rec *r, ezweb_download_conf *conf, table *param)
{
    r->content_type = CONTENT_TYPE_X_UP_DOWNLOAD;
    ap_set_content_length(r, 0);
    ap_send_http_header(r);

    return OK;
}

static int state_success(request_rec *r, ezweb_download_conf *conf, table *param)
{
    r->content_type = CONTENT_TYPE_X_HDML;
    ap_send_http_header(r);
    ap_rprintf(r, HDML_SUCCESS);

    return OK;
}

static int state_fail(request_rec *r, ezweb_download_conf *conf, table *param)
{
    r->content_type = CONTENT_TYPE_X_HDML;
    ap_send_http_header(r);
    ap_rprintf(r, HDML_FAIL);

    return OK;
}

/* content handler */
static int ezweb_download_handler(request_rec *r)
{
    int rc;
    table *param;
    const char *str;
    struct hdml_download_rec dl_param;
    ezweb_download_conf *conf;

    rc = ap_discard_request_body(r);
    if (rc != OK)
        return rc;

    r->allowed |= (1 << M_GET) | (1 << M_OPTIONS);
    if (r->method_number == M_INVALID) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                      "Invalid method in request %s",
                      ap_escape_logitem(r->pool, r->the_request));
        return NOT_IMPLEMENTED;
    }
    if (r->method_number == M_OPTIONS)
        return ap_send_http_options(r);
    if (r->method_number == M_PUT)
        return METHOD_NOT_ALLOWED;
    if (r->finfo.st_mode == 0 || (r->path_info && *r->path_info)) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r,
                      "File does not exist: %s",
                      r->path_info
                         ? ap_pstrcat(r->pool, r->filename, r->path_info, NULL)
                         : r->filename);
        return HTTP_NOT_FOUND;
    }
    if (r->method_number != M_GET)
        return METHOD_NOT_ALLOWED;

    conf = ap_get_module_config(r->per_dir_config, &ezweb_download_module);
    param = parse_request_param(r);
    str = ap_table_get(param, "offset");
    dl_param.offset = str != NULL ? atol(str) : 0;
    str = ap_table_get(param, "count");
    dl_param.count = str != NULL ? atol(str) : 0;
    dl_param.name = ap_table_get(param, "name");

    if (ap_table_get(param, "offset") == NULL && ap_table_get(param, "count") == NULL) {
        rc = state_confirm(r, conf, param);
    }
    else if (dl_param.offset >= 0 && dl_param.count > 0) {
        rc = state_download(r, conf, param);
    }
    else if (dl_param.offset == 0 && dl_param.count == 0) {
        rc = state_empty(r, conf, param);
    }
    else if (dl_param.offset == -1 && dl_param.count == -1) {
        rc = state_success(r, conf, param);
    }
    else if (dl_param.offset == -1 && dl_param.count == -2) {
        rc = state_fail(r, conf, param);
    }
    else {
        rc = state_fail(r, conf, param);
        ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, r,
                      "Unknown request args from UP.Browser: %s",
                      r->unparsed_uri);
    }

    return rc;
}

static const command_rec cmds[] = {
    { "AddEzwebDownloadDatatype", cmd_add_datatype, NULL, OR_FILEINFO, ITERATE2,
      "a disposition data type followed by onen or more file extensions" },
    { NULL }

};

/* Dispatch list of content handlers */
static const handler_rec ezweb_download_handlers[] = { 
    { "ezweb-download", ezweb_download_handler }, 
    { NULL, NULL }
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT ezweb_download_module = {
    STANDARD_MODULE_STUFF, 
    init_module,           /* module initializer                  */
    create_dir_conf,       /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    cmds,                  /* table of config file commands       */
    ezweb_download_handlers,/* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
};

